/*
    LICENSE INFORMATION:
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License (LGPL) as published by the Free Software Foundation.

    Please refer to the COPYING file for more information.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    Copyright © 2004 Bruno T. C. de Oliveira
    Copyright © 2006 Pierre Habouzit
 */

#define _GNU_SOURCE
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <locale.h>
#include <pty.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <wchar.h>

#include "madtty.h"

#define IS_CONTROL(ch) !((ch) & 0xffffff60UL)

enum {
    C0_NUL = 0x00,
            C0_SOH, C0_STX, C0_ETX, C0_EOT, C0_ENQ, C0_ACK, C0_BEL,
    C0_BS , C0_HT , C0_LF , C0_VT , C0_FF , C0_CR , C0_SO , C0_SI ,
    C0_DLE, C0_DC1, C0_DC2, D0_DC3, C0_DC4, C0_NAK, C0_SYN, C0_ETB,
    C0_CAN, C0_EM , C0_SUB, C0_ESC, C0_IS4, C0_IS3, C0_IS2, C0_IS1,
};

enum {
    C1_40 = 0x40,
            C1_41 , C1_BPH, C1_NBH, C1_44 , C1_NEL, C1_SSA, C1_ESA,
    C1_HTS, C1_HTJ, C1_VTS, C1_PLD, C1_PLU, C1_RI , C1_SS2, C1_SS3,
    C1_DCS, C1_PU1, C1_PU2, C1_STS, C1_CCH, C1_MW , C1_SPA, C1_EPA,
    C1_SOS, C1_59 , C1_SCI, C1_CSI, CS_ST , C1_OSC, C1_PM , C1_APC,
};

enum {
    CSI_ICH = 0x40,
             CSI_CUU, CSI_CUD, CSI_CUF, CSI_CUB, CSI_CNL, CSI_CPL, CSI_CHA,
    CSI_CUP, CSI_CHT, CSI_ED , CSI_EL , CSI_IL , CSI_DL , CSI_EF , CSI_EA ,
    CSI_DCH, CSI_SEE, CSI_CPR, CSI_SU , CSI_SD , CSI_NP , CSI_PP , CSI_CTC,
    CSI_ECH, CSI_CVT, CSI_CBT, CSI_SRS, CSI_PTX, CSI_SDS, CSI_SIMD, CSI_5F,
    CSI_HPA, CSI_HPR, CSI_REP, CSI_DA , CSI_VPA, CSI_VPR, CSI_HVP, CSI_TBC,
    CSI_SM , CSI_MC , CSI_HPB, CSI_VPB, CSI_RM , CSI_SGR, CSI_DSR, CSI_DAQ,
    CSI_70 , CSI_71 , CSI_72 , CSI_73 , CSI_74 , CSI_75 , CSI_76 , CSI_77 ,
    CSI_78 , CSI_79 , CSI_7A , CSI_7B , CSI_7C , CSI_7D , CSI_7E , CSI_7F
};

struct mtty_row_t {
    wchar_t  *text;
    uint16_t *attr;
    unsigned dirty : 1;
};

static char const * const keytable[KEY_MAX+1] = {
    ['\n']          = "\r",
    [KEY_UP]        = "\e[A",
    [KEY_DOWN]      = "\e[B",
    [KEY_RIGHT]     = "\e[C",
    [KEY_LEFT]      = "\e[D",
    [KEY_BACKSPACE] = "\177",
    [KEY_HOME]      = "\e[1~",
    [KEY_IC]        = "\e[2~",
    [KEY_DC]        = "\e[3~",
    [KEY_END]       = "\e[4~",
    [KEY_PPAGE]     = "\e[5~",
    [KEY_NPAGE]     = "\e[6~",
    [KEY_SUSPEND]   = "\x1A",  /* Ctrl+Z gets mapped to this */
    [KEY_F(1)]      = "\e[[A",
    [KEY_F(2)]      = "\e[[B",
    [KEY_F(3)]      = "\e[[C",
    [KEY_F(4)]      = "\e[[D",
    [KEY_F(5)]      = "\e[[E",
    [KEY_F(6)]      = "\e[17~",
    [KEY_F(7)]      = "\e[18~",
    [KEY_F(8)]      = "\e[19~",
    [KEY_F(9)]      = "\e[20~",
    [KEY_F(10)]     = "\e[21~",
};

static void mtty_row_set(mtty_row_t *row, int start, int len, uint16_t attr)
{
    row->dirty = true;
    wmemset(row->text + start, 0, len);
    for (int i = start; i < len + start; i++) {
        row->attr[i] = attr;
    }
}

static void mtty_row_roll(mtty_row_t *start, mtty_row_t *end, int count)
{
    int n = end - start;

    count %= n;
    if (count < 0)
        count += n;

    if (count) {
        mtty_row_t *buf = alloca(count * sizeof(mtty_row_t));

        memcpy(buf, start, count * sizeof(mtty_row_t));
        memmove(start, start + count, (n - count) * sizeof(mtty_row_t));
        memcpy(end - count, buf, count * sizeof(mtty_row_t));
        for (mtty_row_t *row = start; row < end; row++) {
            row->dirty = true;
        }
    }
}

static void clamp_cursor_to_bounds(madtty_t *rt)
{
    if (rt->curs_row < rt->lines) {
        rt->curs_row = rt->lines;
    }
    if (rt->curs_row >= rt->lines + rt->rows) {
        rt->curs_row = rt->lines + rt->rows - 1;
    }

    if (rt->curs_col < 0) {
        rt->curs_col = 0;
    }
    if (rt->curs_col >= rt->cols) {
        rt->curs_col = rt->cols - 1;
    }
}

static void cursor_line_down(madtty_t *rt)
{
    rt->curs_row++;
    if (rt->curs_row < rt->scroll_bot)
        return;

    rt->curs_row = rt->scroll_bot - 1;
    mtty_row_roll(rt->scroll_top, rt->scroll_bot, 1);
    mtty_row_set(rt->curs_row, 0, rt->cols, 0);
}

__attribute__((const))
static uint16_t build_attrs(unsigned curattrs)
{
    return ((curattrs & ~A_COLOR) | COLOR_PAIR(curattrs & 0xff))
        >> NCURSES_ATTR_SHIFT;
}

static void new_escape_sequence(madtty_t *rt)
{
    rt->escaped = true;
    rt->elen    = 0;
    rt->ebuf[0] = '\0';
}

static void cancel_escape_sequence(madtty_t *rt)
{
    rt->escaped = false;
    rt->elen    = 0;
    rt->ebuf[0] = '\0';
}

static bool is_valid_csi_ender(int c)
{
    return (c >= 'a' && c <= 'z')
        || (c >= 'A' && c <= 'Z')
        || (c == '@' || c == '`');
}

/* interprets a 'set attribute' (SGR) CSI escape sequence */
static void interpret_csi_SGR(madtty_t *rt, int param[], int pcount)
{
    int i;

    if (pcount == 0) {
        /* special case: reset attributes */
        rt->curattrs = A_NORMAL;
        return;
    }

    for (i = 0; i < pcount; i++) {
        switch (param[i]) {
#define CASE(x, op)  case x: op; break
            CASE(0,  rt->curattrs = A_NORMAL);
            CASE(1,  rt->curattrs |= A_BOLD);
            CASE(4,  rt->curattrs |= A_UNDERLINE);
            CASE(5,  rt->curattrs |= A_BLINK);
            CASE(6,  rt->curattrs |= A_BLINK);
            CASE(7,  rt->curattrs |= A_REVERSE);
            CASE(8,  rt->curattrs |= A_INVIS);
            CASE(22, rt->curattrs &= ~A_BOLD);
            CASE(24, rt->curattrs &= ~A_UNDERLINE);
            CASE(25, rt->curattrs &= ~A_BLINK);
            CASE(27, rt->curattrs &= ~A_REVERSE);
            CASE(28, rt->curattrs &= ~A_INVIS);

          case 30 ... 37:
            rt->curattrs &= ~0xf0;
            rt->curattrs |= (param[i] - 29) << 4;
            break;

          case 39:
            rt->curattrs &= ~0xf0;
            break;

          case 40 ... 47:
            rt->curattrs &= ~0x0f;
            rt->curattrs |= (param[i] - 39);
            break;

          case 49:
            rt->curattrs &= ~0x0f;
            break;

          default:
            break;
        }
    }
}

/* interprets an 'erase display' (ED) escape sequence */
static void interpret_csi_ED(madtty_t *rt, int param[], int pcount)
{
    mtty_row_t *row, *start, *end;
    attr_t attr = build_attrs(rt->curattrs);

    /* decide range */
    if (pcount && param[0] == 2) {
        start = rt->lines;
        end   = rt->lines + rt->rows;
    } else
    if (pcount && param[0] == 1) {
        start = rt->lines;
        end   = rt->curs_row;
        mtty_row_set(rt->curs_row, 0, rt->curs_col + 1, attr);
    } else {
        mtty_row_set(rt->curs_row, rt->curs_col,
                     rt->cols - rt->curs_col, attr);
        start = rt->curs_row + 1;
        end   = rt->lines + rt->rows;
    }

    for (row = start; row < end; row++) {
        mtty_row_set(row, 0, rt->cols, attr);
    }
}

/* interprets a 'move cursor' (CUP) escape sequence */
static void interpret_csi_CUP(madtty_t *rt, int param[], int pcount)
{
    if (pcount == 0) {
        /* special case */
        rt->curs_row = rt->lines;
        rt->curs_col = 0;
        return;
    } else
    if (pcount < 2) {
        return;  /* malformed */
    }

    rt->curs_row = rt->lines + param[0] - 1;  /* convert from 1-based to 0-based */
    rt->curs_col = param[1] - 1;  /* convert from 1-based to 0-based */

    clamp_cursor_to_bounds(rt);
}

/* Interpret the 'relative mode' sequences: CUU, CUD, CUF, CUB, CNL,
 * CPL, CHA, HPR, VPA, VPR, HPA */
static void interpret_csi_C(madtty_t *rt, char verb, int param[], int pcount)
{
    int n = (pcount && param[0] > 0) ? param[0] : 1;

    switch (verb) {
      case 'A':           rt->curs_row -= n; break;
      case 'B': case 'e': rt->curs_row += n; break;
      case 'C': case 'a': rt->curs_col += n; break;
      case 'D':           rt->curs_col -= n; break;
      case 'E':           rt->curs_row += n; rt->curs_col = 0; break;
      case 'F':           rt->curs_row -= n; rt->curs_col = 0; break;
      case 'G': case '`': rt->curs_col  = param[0] - 1; break;
      case 'd':           rt->curs_row  = rt->lines + param[0] - 1; break;
    }

    clamp_cursor_to_bounds(rt);
}

/* Interpret the 'erase line' escape sequence */
static void interpret_csi_EL(madtty_t *rt, int param[], int pcount)
{
    attr_t attr = build_attrs(rt->curattrs);

    switch (pcount ? param[0] : 0) {
      case 1:
        mtty_row_set(rt->curs_row, 0, rt->curs_col + 1, attr);
        break;
      case 2:
        mtty_row_set(rt->curs_row, 0, rt->cols, attr);
        break;
      default:
        mtty_row_set(rt->curs_row, rt->curs_col, rt->cols - rt->curs_col,
                     attr);
        break;
    }
}

/* Interpret the 'insert blanks' sequence (ICH) */
static void interpret_csi_ICH(madtty_t *rt, int param[], int pcount)
{
    mtty_row_t *row = rt->curs_row;
    int n = (pcount && param[0] > 0) ? param[0] : 1;
    int i;

    if (rt->curs_col + n > rt->cols) {
        n = rt->cols - rt->curs_col;
    }

    for (i = rt->cols - 1; i >= rt->curs_col + n; i--) {
        row->text[i] = row->text[i - n];
        row->attr[i] = row->attr[i - n];
    }

    mtty_row_set(row, rt->curs_col, n, build_attrs(rt->curattrs));
}

/* Interpret the 'delete chars' sequence (DCH) */
static void interpret_csi_DCH(madtty_t *rt, int param[], int pcount)
{
    mtty_row_t *row = rt->curs_row;
    int n = (pcount && param[0] > 0) ? param[0] : 1;
    int i;

    if (rt->curs_col + n > rt->cols) {
        n = rt->cols - rt->curs_col;
    }

    for (i = rt->curs_col; i < rt->cols - n; i++) {
        row->text[i] = row->text[i + n];
        row->attr[i] = row->attr[i + n];
    }

    mtty_row_set(row, rt->cols - n, n, build_attrs(rt->curattrs));
}

/* Interpret a 'scroll reverse' (SR) */
static void interpret_csi_SR(madtty_t *rt)
{
    mtty_row_roll(rt->scroll_top, rt->scroll_bot, -1);
    mtty_row_set(rt->scroll_top, 0, rt->cols, build_attrs(rt->curattrs));
}

/* Interpret an 'insert line' sequence (IL) */
static void interpret_csi_IL(madtty_t *rt, int param[], int pcount)
{
    int n = (pcount && param[0] > 0) ? param[0] : 1;

    if (rt->curs_row + n >= rt->scroll_bot) {
        for (mtty_row_t *row = rt->curs_row; row < rt->scroll_bot; row++) {
            mtty_row_set(row, 0, rt->cols, build_attrs(rt->curattrs));
        }
    } else {
        mtty_row_roll(rt->curs_row, rt->scroll_bot, -n);
        for (mtty_row_t *row = rt->curs_row; row < rt->curs_row + n; row++) {
            mtty_row_set(row, 0, rt->cols, build_attrs(rt->curattrs));
        }
    }
}

/* Interpret a 'delete line' sequence (DL) */
static void interpret_csi_DL(madtty_t *rt, int param[], int pcount)
{
    int n = (pcount && param[0] > 0) ? param[0] : 1;

    if (rt->curs_row + n >= rt->scroll_bot) {
        for (mtty_row_t *row = rt->curs_row; row < rt->scroll_bot; row++) {
            mtty_row_set(row, 0, rt->cols, build_attrs(rt->curattrs));
        }
    } else {
        mtty_row_roll(rt->curs_row, rt->scroll_bot, n);
        for (mtty_row_t *row = rt->scroll_bot - n; row < rt->scroll_bot; row++) {
            mtty_row_set(row, 0, rt->cols, build_attrs(rt->curattrs));
        }
    }
}

/* Interpret an 'erase characters' (ECH) sequence */
static void interpret_csi_ECH(madtty_t *rt, int param[], int pcount)
{
    int n = (pcount && param[0] > 0) ? param[0] : 1;

    if (rt->curs_col + n < rt->cols) {
        n = rt->cols - rt->curs_col;
    }
    mtty_row_set(rt->curs_row, rt->curs_col, n, build_attrs(rt->curattrs));
}

/* Interpret a 'set scrolling region' (DECSTBM) sequence */
static void interpret_csi_DECSTBM(madtty_t *rt, int param[], int pcount)
{
    int new_top, new_bot;

    switch (pcount) {
      case 0:
        rt->scroll_top = rt->lines;
        rt->scroll_bot = rt->lines + rt->rows;
        break;
      default:
        return; /* malformed */

      case 2:
        new_top = param[0] - 1;
        new_bot = param[1];

        /* clamp to bounds */
        if (new_top < 0)
            new_top = 0;
        if (new_top >= rt->rows)
            new_top = rt->rows - 1;
        if (new_bot < 0)
            new_bot = 0;
        if (new_bot >= rt->rows)
            new_bot = rt->rows;

        /* check for range validity */
        if (new_top < new_bot) {
            rt->scroll_top = rt->lines + new_top;
            rt->scroll_bot = rt->lines + new_bot;
        }
        break;
    }
}

static void es_interpret_csi(madtty_t *rt)
{
    static int csiparam[MAX_CSI_ES_PARAMS];
    int param_count = 0;
    const char *p = rt->ebuf + 1;
    char verb = rt->ebuf[rt->elen - 1];

    p += rt->ebuf[1] == '?'; /* CSI private mode */

    /* parse numeric parameters */
    while (isdigit((unsigned char)*p) || *p == ';') {
        if (*p == ';') {
            if (param_count >= MAX_CSI_ES_PARAMS) return; /* too long! */
            csiparam[param_count++] = 0;
        } else {
            if (param_count == 0) csiparam[param_count++] = 0;
            csiparam[param_count - 1] *= 10;
            csiparam[param_count - 1] += *p - '0';
        }

        p++;
    }

    if (rt->ebuf[1] == '?') {
        switch (verb) {
          case 'l':
            if (csiparam[0] == 25)
                rt->curshid = true;
            break;

          case 'h':
            if (csiparam[0] == 25)
                rt->curshid = false;
            break;
        }
    }

    /* delegate handling depending on command character (verb) */
    switch (verb) {
      case 'h':
        if (param_count == 1 && csiparam[0] == 4) /* insert mode */
            rt->insert = true;
        break;
      case 'l':
        if (param_count == 1 && csiparam[0] == 4) /* replace mode */
            rt->insert = false;
        break;
      case 'm': /* it's a 'set attribute' sequence */
        interpret_csi_SGR(rt, csiparam, param_count); break;
      case 'J': /* it's an 'erase display' sequence */
        interpret_csi_ED(rt, csiparam, param_count); break;
      case 'H': case 'f': /* it's a 'move cursor' sequence */
        interpret_csi_CUP(rt, csiparam, param_count); break;
      case 'A': case 'B': case 'C': case 'D': case 'E': case 'F': case 'G':
      case 'e': case 'a': case 'd': case '`':
        /* it is a 'relative move' */
        interpret_csi_C(rt, verb, csiparam, param_count); break;
      case 'K': /* erase line */
        interpret_csi_EL(rt, csiparam, param_count); break;
      case '@': /* insert characters */
        interpret_csi_ICH(rt, csiparam, param_count); break;
      case 'P': /* delete characters */
        interpret_csi_DCH(rt, csiparam, param_count); break;
      case 'L': /* insert lines */
        interpret_csi_IL(rt, csiparam, param_count); break;
      case 'M': /* delete lines */
        interpret_csi_DL(rt, csiparam, param_count); break;
      case 'X': /* erase chars */
        interpret_csi_ECH(rt, csiparam, param_count); break;
      case 'r': /* set scrolling region */
        interpret_csi_DECSTBM(rt, csiparam, param_count); break;
      case 's': /* save cursor location */
        rt->curs_srow = rt->curs_row - rt->lines;
        rt->curs_scol = rt->curs_col;
        break;
      case 'u': /* restore cursor location */
        rt->curs_row = rt->lines + rt->curs_srow;
        rt->curs_col = rt->curs_scol;
        clamp_cursor_to_bounds(rt);
        break;
      default:
        break;
    }
}

static void try_interpret_escape_seq(madtty_t *rt)
{
    char lastchar  = rt->ebuf[rt->elen-1];

    switch (*rt->ebuf) {
      case '\0':
        return;

      case 'M':
        interpret_csi_SR(rt);
        cancel_escape_sequence(rt);
        return;

      case '(':
      case ')':
        if (rt->elen == 2)
            goto cancel;
        break;

      case ']': /* xterm thing */
        if (lastchar == '\a')
            goto cancel;
        break;

      default:
        goto cancel;

      case '[':
        if (is_valid_csi_ender(lastchar)) {
            es_interpret_csi(rt);
            cancel_escape_sequence(rt);
            return;
        }
        break;
    }

    if (rt->elen + 1 >= (int)sizeof(rt->ebuf)) {
cancel:
#if 0
        int i;
        fprintf(stderr, "cancelled: \\033");
        for (i = 0; i < rt->elen; i++) {
            int c = rt->ebuf[i];
            if (isprint(c) && c >= ' ') {
                fputc(c, stderr);
            } else {
                fprintf(stderr, "\\%03o", c);
            }
        }
        fputc('\n', stderr);
#endif
        cancel_escape_sequence(rt);
    }
}

static void madtty_process_nonprinting(madtty_t *rt, wchar_t wc)
{
    switch (wc) {
      case C0_ESC:
        new_escape_sequence(rt);
        break;

      case C0_BEL:
        /* do nothing for now... maybe a visual bell would be nice? */
        break;

      case C0_BS:
        if (rt->curs_col > 0)
            rt->curs_col--;
        break;

      case C0_HT: /* tab */
        rt->curs_col = (rt->curs_col + 8) & ~7;
        if (rt->curs_col >= rt->cols)
            rt->curs_col = rt->cols - 1;
        break;

      case C0_CR:
        rt->curs_col = 0;
        break;

      case C0_VT:
      case C0_FF:
      case C0_LF:
        cursor_line_down(rt);
        break;

      case C0_SO:		/* shift out - acs */
        rt->graphmode = true;
        break;
      case C0_SI:		/* shift in - acs */
        rt->graphmode = false;
        break;
    }
}

void madtty_putc(madtty_t *rt, wchar_t wc)
{
    int width = 0;

    if (!rt->seen_input) {
        rt->seen_input = 1;
        kill(-rt->childpid, SIGWINCH);
    }

#if 0
    if (wc == '\n' || (wc >= ' ' && isprint(wc))) {
        fputc(wc, stderr);
    } else {
        fprintf(stderr, "\\%03o", wc);
    }
#endif

    if (rt->escaped) {
        assert (rt->elen + 1 < (int)sizeof(rt->ebuf));
        rt->ebuf[rt->elen]   = wc;
        rt->ebuf[++rt->elen] = '\0';
        try_interpret_escape_seq(rt);
    } else if (IS_CONTROL(wc)) {
        madtty_process_nonprinting(rt, wc);
    } else {
        mtty_row_t *tmp;

        if (rt->graphmode) {
            // vt100 special graphics and line drawing
            // 5f-7e standard vt100
            // 40-5e rxvt extension for extra curses acs chars
            static uint16_t vt100_0[62] = { // 41 .. 7e
                        0x2191, 0x2193, 0x2192, 0x2190, 0x2588, 0x259a, 0x2603, // 41-47 hi mr. snowman!
                     0,      0,      0,      0,      0,      0,      0,      0, // 48-4f
                     0,      0,      0,      0,      0,      0,      0,      0, // 50-57
                     0,      0,      0,      0,      0,      0,      0, 0x0020, // 58-5f
                0x25c6, 0x2592, 0x2409, 0x240c, 0x240d, 0x240a, 0x00b0, 0x00b1, // 60-67
                0x2424, 0x240b, 0x2518, 0x2510, 0x250c, 0x2514, 0x253c, 0x23ba, // 68-6f
                0x23bb, 0x2500, 0x23bc, 0x23bd, 0x251c, 0x2524, 0x2534, 0x252c, // 70-77
                0x2502, 0x2264, 0x2265, 0x03c0, 0x2260, 0x00a3, 0x00b7,         // 78-7e
            };

            if (wc >= 0x41 && wc <= 0x7e && vt100_0[wc - 0x41]) {
                wc = vt100_0[wc - 0x41];
            }
            width = 1;
        } else {
            width = wcwidth(wc) ?: 1;
        }

        if (width == 2 && rt->curs_col == rt->cols - 1) {
            tmp = rt->curs_row;
            tmp->dirty = true;
            tmp->text[rt->curs_col] = 0;
            tmp->attr[rt->curs_col] = build_attrs(rt->curattrs);
            rt->curs_col++;
        }

        if (rt->curs_col >= rt->cols) {
            rt->curs_col = 0;
            cursor_line_down(rt);
        }

        tmp = rt->curs_row;
        tmp->dirty = true;

        if (rt->insert) {
            wmemmove(tmp->text + rt->curs_col + width, tmp->text + rt->curs_col,
                     (rt->cols - rt->curs_col - width));
            memmove(tmp->attr + rt->curs_col + width, tmp->attr + rt->curs_col,
                    (rt->cols - rt->curs_col - width) * sizeof(tmp->attr[0]));
        }

        tmp->text[rt->curs_col] = wc;
        tmp->attr[rt->curs_col] = build_attrs(rt->curattrs);
        rt->curs_col++;
        if (width == 2) {
            tmp->text[rt->curs_col] = 0;
            tmp->attr[rt->curs_col] = build_attrs(rt->curattrs);
            rt->curs_col++;
        }
    }
}

int madtty_process(madtty_t *rt)
{
    int res, pos = 0;

    if (rt->pty < 0) {
        errno = EINVAL;
        return -1;
    }

    res = read(rt->pty, rt->rbuf + rt->rlen, sizeof(rt->rbuf) - rt->rlen);
    if (res < 0)
        return -1;

    rt->rlen += res;
    while (pos < rt->rlen) {
        wchar_t wc;
        ssize_t len;

        len = (ssize_t)mbrtowc(&wc, rt->rbuf + pos, rt->rlen - pos, &rt->ps);
        if (len == -2) {
            rt->rlen -= pos;
            memmove(rt->rbuf, rt->rbuf + pos, rt->rlen);
            return 0;
        }

        if (len == -1) {
            len = 1;
            wc  = rt->rbuf[pos];
        }

        pos += len ? len : 1;
        madtty_putc(rt, wc);
    }

    rt->rlen -= pos;
    memmove(rt->rbuf, rt->rbuf + pos, rt->rlen);
    return 0;
}

madtty_t *madtty_create(int rows, int cols)
{
    madtty_t *rt;
    int i;

    if (rows <= 0 || cols <= 0)
        return NULL;

    rt = (madtty_t*)calloc(sizeof(madtty_t), 1);
    if (!rt)
        return NULL;

    /* record dimensions */
    rt->rows = rows;
    rt->cols = cols;

    /* default mode is replace */
    rt->insert = false; 

    /* create the cell matrix */
    rt->lines = (mtty_row_t*)calloc(sizeof(mtty_row_t), rt->rows);
    for (i = 0; i < rt->rows; i++) {
        rt->lines[i].text = (wchar_t *)calloc(sizeof(wchar_t), rt->cols);
        rt->lines[i].attr = (uint16_t *)calloc(sizeof(uint16_t), rt->cols);
    }

    rt->pty = -1;  /* no pty for now */

    /* initialization of other public fields */
    rt->curs_row = rt->lines;
    rt->curs_col = 0;
    rt->curattrs = A_NORMAL;  /* white text over black background */

    /* initial scrolling area is the whole window */
    rt->scroll_top = rt->lines;
    rt->scroll_bot = rt->lines + rt->rows;

    return rt;
}

void madtty_destroy(madtty_t *rt)
{
    int i;
    if (!rt)
        return;

    for (i = 0; i < rt->rows; i++) {
        free(rt->lines[i].text);
        free(rt->lines[i].attr);
    }
    free(rt->lines);
    free(rt);
}

void madtty_draw(madtty_t *rt, WINDOW *win, int srow, int scol)
{
    curs_set(0);
    for (int i = 0; i < rt->rows; i++) {
        mtty_row_t *row = rt->lines + i;


        if (!row->dirty)
            continue;

        wmove(win, srow + i, scol);
        for (int j = 0; j < rt->cols; j++) {
            if (!j || row->attr[j] != row->attr[j - 1])
                wattrset(win, (attr_t)row->attr[j] << NCURSES_ATTR_SHIFT);
            if (row->text[j] >= 128) {
                char buf[MB_CUR_MAX + 1];
                int len;

                len = wcrtomb(buf, row->text[j], NULL);
                waddnstr(win, buf, len);
                if (wcwidth(row->text[j]) > 1)
                    j++;
            } else {
                waddch(win, row->text[j] > ' ' ? row->text[j] : ' ');
            }
        }
        row->dirty = false;
    }

    wmove(win, srow + rt->curs_row - rt->lines, scol + rt->curs_col);
    curs_set(!rt->curshid);
}

/******************************************************/

pid_t madtty_forkpty(madtty_t *rt, const char *path, const char *argv[])
{
    struct winsize ws;
    pid_t pid;

    ws.ws_row    = rt->rows;
    ws.ws_col    = rt->cols;
    ws.ws_xpixel = ws.ws_ypixel = 0;

    pid = forkpty(&rt->pty, NULL, NULL, &ws);
    if (pid < 0)
        return -1;

    if (pid == 0) {
        setsid();
        setenv("TERM", "rxvt", 1);
        execv(path, (char *const*)argv);
        fprintf(stderr, "\nexecv() failed.\nCommand: '%s'\n", argv[0]);
        exit(1);
    }

    return rt->childpid = pid;
}

void madtty_keypress(madtty_t *rt, int keycode)
{
    char c = (char)keycode;
    const char *buf;
    int len;

#if 0
    if (keycode == KEY_F(1)) {
#define MIN(a, b) ((a < (b)) ? a : (b))
        kill(-rt->childpid, SIGWINCH);
        rt->scroll_bot = MIN(rt->scroll_bot, rt->lines + rt->rows);
        rt->curs_row = MIN(rt->curs_row, rt->lines + rt->rows);
        printf(stderr, "%d\n", rt->rows);
        return;
    }
#endif
    if (keycode >= 0 && keycode < KEY_MAX && keytable[keycode]) {
        buf = keytable[keycode];
        len = strlen(keytable[keycode]);
    } else {
        buf = &c;
        len = 1;
    }

    while (len > 0) {
        int res = write(rt->pty, buf, len);
        if (res < 0 && errno != EAGAIN && errno != EINTR)
            return;

        buf += res;
        len -= res;
    }
}

void madtty_initialize(void)
{
    setlocale(LC_ALL, "");
    initscr();
    start_color();
    use_default_colors();
    noecho();
    raw();
    nodelay(stdscr, TRUE);
    keypad(stdscr, TRUE);
    ESCDELAY=50;

    for (int i = -1; i < 8; i++) {
        for (int j = -1; j < 8; j++) {
            init_pair((i + 1) * 16 + j + 1, i, j);
        }
    }
}
